#! /usr/bin/env python
"""
Example usage:
    snscrape http://csdms.colorado.edu/wiki/CSN_Quantity_Templates \
             http://csdms.colorado.edu/wiki/CSN_Object_Templates \
             http://csdms.colorado.edu/wiki/CSN_Operation_Templates \
            > data/scraped.yaml
"""
from __future__ import print_function

import os

from ..utilities import FORMATTERS, SCRAPERS, scrape


_AS_TXT = FORMATTERS["txt"]

_DEFAULT_SEARCH = r"\b[\w~-]+__[\w~-]+"


def snscrape(files, with_headers=False, regex=None, format="url", newline=None):
    """Scrape names from a URL.

    Parameters
    ----------
    files : iterable of str
        List of files or URL to scrape.
    with_headers : bool, optional
        Include headers in the output that indicate the name of the source.
    regex : str, optional
        A regular expression that defines what a Standard Name is.
    format : {'url', 'plain_text'}, optional
        The format of the target that's being scraped.
    newline : str, optional
        Newline character to use for output.

    Returns
    -------
    str
        The scraped names.

    Examples
    --------
    >>> from __future__ import print_function
    >>> from six.moves import StringIO
    >>> import standard_names as csn

    >>> file1 = StringIO(\"\"\"
    ... A file is one name, which is air__temperature.
    ... \"\"\")
    >>> file2 = StringIO(\"\"\"
    ... A file is two names: air__temperature, and water__temperature.
    ... \"\"\")

    >>> lines = csn.cmd.snscrape.snscrape([file1, file2], format='plain_text')
    >>> sorted(lines.split(os.linesep))
    ['air__temperature', 'air__temperature', 'water__temperature']
    """
    newline = newline or os.linesep
    regex = regex or _DEFAULT_SEARCH

    docs = {}
    for file_name in files:
        docs[file_name] = scrape(file_name, regex=regex, format=format)

    documents = []
    for (name, name_list) in docs.items():
        if with_headers:
            heading = "Scraped from %s" % name
        else:
            heading = None
        documents.append(_AS_TXT(name_list, sorted=True, heading=heading))

    return newline.join(documents)


def main(args=None):
    """Scrape standard names from a file or URL.

    Examples
    --------
    >>> import os
    >>> import tempfile
    >>> import standard_names as csn

    >>> contents = \"\"\"
    ... A file with text and names (air__temperature) mixed in. Some names
    ... have double underscores (like, Water__Temperature) by are not
    ... valid names. Others, like water__temperature, are good.
    ... \"\"\"

    >>> (fd, fname) = tempfile.mkstemp()
    >>> os.close(fd)

    >>> with open(fname, 'w') as fp:
    ...     print(contents, file=fp)

    >>> names = csn.cmd.snscrape.main(
    ...     [fp.name, '--reader=plain_text', '--no-headers'])
    >>> names.split(os.linesep)
    ['air__temperature', 'water__temperature']

    >>> os.remove(fname)
    """
    import argparse

    parser = argparse.ArgumentParser("Scrape standard names from a file or URL")
    parser.add_argument("file", nargs="+", metavar="FILE", help="URL or file to scrape")
    parser.add_argument(
        "--reader", choices=SCRAPERS.keys(), default="url", help="Name of reader"
    )
    parser.add_argument(
        "--regex",
        default=_DEFAULT_SEARCH,
        help="Regular expression describing " "a standard name (%s)" % _DEFAULT_SEARCH,
    )
    parser.add_argument(
        "--no-headers", action="store_true", help="Do not print headers between scrapes"
    )

    if args is None:
        args = parser.parse_args()
    else:
        args = parser.parse_args(args)

    return snscrape(
        args.file,
        with_headers=not args.no_headers,
        regex=args.regex,
        format=args.reader,
    )


def run():
    print(main())



"""The CSDMS Standard Names"""

from .standardname import StandardName, is_valid_name
from .registry import NamesRegistry
from .error import BadNameError, BadRegistryError


from ._version import get_versions
__version__ = get_versions()['version']
del get_versions



#! /usr/bin/env python
"""Some IO functions for standard_names package."""
from __future__ import print_function

import os
import sys
import re
import warnings

from ..standardname import StandardName
from ..registry import NamesRegistry
from ..error import BadNameError
from .decorators import (
    format_as_wiki,
    format_as_yaml,
    format_as_plain_text,
    google_doc,
    url,
    plain_text,
)


def _list_to_string(lines, **kwds):
    """Join strings with the line separator.

    Concatonate a list of strings into one big string using the line separator
    as a joiner.

    Parameters
    ----------
    lines : iterable of str
        String to join.
    sorted : bool, optional
        Sort the strings before joining them.
    newline : str, optional
        Newline character to use for output.

    Returns
    -------
    str
        The joined strings.

    Examples
    --------
    >>> from __future__ import print_function
    >>> import standard_names as csn
    >>> print(csn.utilities.io._list_to_string(('foo', 'bar'), newline='\\n'))
    foo
    bar
    >>> print(csn.utilities.io._list_to_string(('foo', 'bar'), sorted=True, newline='\\n'))
    bar
    foo
    """
    newline = kwds.pop("newline", os.linesep)
    if kwds.pop("sorted", False):
        sorted_lines = list(lines)
        sorted_lines.sort()
        return newline.join(sorted_lines)
    else:
        return newline.join(lines)


def _scrape_stream(stream, regex=r"\b\w+__\w+"):
    """Scrape standard names from stream matching a regular expression.

    Parameters
    ----------
    stream : file_like
        File-like object from which to read (only a ``read`` method is
        necessary).
    regex : str, optional
        A regular expression that indicates a word to scrape.

    Returns
    -------
    NamesRegistry
        The scraped words.

    Examples
    --------
    >>> import standard_names as csn
    >>> from six.moves import StringIO
    >>> stream = StringIO(\"\"\"
    ... Some text with a standard name (air__temperature) in it.
    ... More words with more names: water__temperature. If a word matches
    ... the pattern but is not a valid name, ignore it (Air__Temperature
    ... is an example).
    ... \"\"\")
    >>> names = csn.utilities.io._scrape_stream(stream)
    >>> sorted(names.names)
    ['air__temperature', 'water__temperature']
    """
    names = NamesRegistry(None)

    text = stream.read()
    words = re.findall(regex, text)
    for word in words:
        try:
            names.add(word)
        except BadNameError as error:
            print(
                "{name}: matches pattern but not a valid name. "
                "Ignoring.".format(name=error.name),
                file=sys.stderr,
            )

    return names


FORMATTERS = {
    "plain": _list_to_string,
    "wiki": format_as_wiki(_list_to_string),
    "yaml": format_as_yaml(_list_to_string),
    "txt": format_as_plain_text(_list_to_string),
}


SCRAPERS = dict()
for decorator in [google_doc, url, plain_text]:
    SCRAPERS[decorator.__name__] = decorator(_scrape_stream)


_VALID_INTENTS = ["input", "output"]


def _find_unique_names(models):
    """Find unique names in a iterable of StandardNames.

    Parameters
    ----------
    models : dict
        Dictionary of model information

    Returns
    -------
    NamesRegistry
        A collection of unique names.
    """
    names = NamesRegistry(None)
    for model in models:
        if isinstance(model["exchange items"], dict):
            new_names = []
            for intent in model["exchange items"]:
                if intent not in _VALID_INTENTS:
                    raise ValueError("{intent}: Bad intent".format(intent=intent))
                new_names.extend(model["exchange items"][intent])
        else:
            new_names = model["exchange items"]

        for new_name in new_names:
            names.add(new_name)
            # names.add(StandardName(new_name))

    return names


def from_model_file(stream):
    """Read names from a model file.

    Get standard names from a YAML file listing standard names for particular
    models and produce the corresponding NamesRegistry.

    Parameters
    ----------
    stream : file_like
        YAML stream.

    Returns
    -------
    NamesRegistry
        A collection of names for the model file.
    """
    import yaml

    models = yaml.load_all(stream)
    names = _find_unique_names(models)
    return names


def from_list_file(stream):
    """Read names from a text file.

    Parameters
    ----------
    stream : file_like
        Source from which to read names (requires only a ``readline`` method).

    Returns
    -------
    NamesRegistry
        A collection of names read from the source.

    Examples
    --------
    >>> import standard_names as csn
    >>> from six.moves import StringIO
    >>> stream = StringIO(\"\"\"
    ... air__temperature
    ... # A comment
    ... water__temperature # Another comment
    ... \"\"\")
    >>> names = csn.utilities.from_list_file(stream)
    >>> sorted(names.names)
    ['air__temperature', 'water__temperature']
    """
    names = NamesRegistry(None)
    for line in stream:
        if "#" in line:
            line = line[: line.find("#")]
        line = line.strip()
        if line:
            names.add(line)
    return names


def scrape(source, **kwds):
    """Scrape standard names for a named source.

    Parameters
    ----------
    source : str
        Name of the source.
    format: str, optional
        The format of the source. Valid values are given by the keys
        of the ``SCRAPERS` global. Currently this is ``google_doc``,
        ``url``, and ``plain_text`` (default is ``url``).

    Returns
    -------
    NamesRegistry
        A collection of names read from the source.
    """
    source_format = kwds.pop("format", "url")

    return SCRAPERS[source_format](source, **kwds)



#! /usr/bin/env python
"""
Example usage:
    snbuild data/models.yaml data/scraped.yaml \
            > standard_names/data/standard_names.yaml
"""
from __future__ import print_function

import os
from ..utilities.io import FORMATTERS
from ..registry import NamesRegistry


def snbuild(file, newline=None):
    """Build a YAML-formatted database of names.

    Parameters
    ----------
    file : str
        Text file of names.
    newline : str, optional
        Newline character to use for output.

    Returns
    -------
    str
        YAML-formatted text of the names database.

    Examples
    --------
    >>> from __future__ import print_function
    >>> import os
    >>> from six.moves import StringIO
    >>> import standard_names as csn

    >>> lines = os.linesep.join(['air__temperature', 'water__temperature'])
    >>> names = StringIO(lines)

    >>> print(csn.cmd.snbuild.snbuild(names, newline='\\n'))
    %YAML 1.2
    ---
    names:
      - air__temperature
      - water__temperature
    ---
    objects:
      - air
      - water
    ---
    quantities:
      - temperature
    ---
    operators:
      []
    ...
    """
    newline = newline or os.linesep
    names = NamesRegistry(file)

    formatter = FORMATTERS["yaml"]

    lines = [
        "%YAML 1.2",
        "---",
        formatter(names.names, sorted=True, heading="names", newline=newline),
        "---",
        formatter(names.objects, sorted=True, heading="objects", newline=newline),
        "---",
        formatter(names.quantities, sorted=True, heading="quantities", newline=newline),
        "---",
        formatter(names.operators, sorted=True, heading="operators", newline=newline),
        "...",
    ]
    return newline.join(lines)


def main(args=None):
    """Build a list of CSDMS standard names for YAML description files."""
    import argparse

    parser = argparse.ArgumentParser(
        "Scan a model description file for CSDMS standard names"
    )
    parser.add_argument(
        "file",
        nargs="+",
        type=argparse.FileType("r"),
        help="YAML file describing model exchange items",
    )

    if args is None:
        args = parser.parse_args()
    else:
        args = parser.parse_args(args)

    return snbuild(args.file)


def run():
    print(main())



#! /usr/bin/env python
from __future__ import print_function

import os

from ..utilities import FORMATTERS
from ..registry import NamesRegistry


_NAMES_SCHEMA = """
create table names (
    id       integer primary key,
    name     text,
    unique(name)
);
create table objects (
    id       integer primary key,
    name     text,
    unique(name)
);
create table quantities (
    id       integer primary key,
    name     text,
    unique(name)
);
create table operators (
    id       integer primary key,
    name     text,
    unique(name)
);
""".strip()


def as_sql_commands(names, newline=None):
    """Create an sql database from a NamesRegistry.

    Parameters
    ----------
    names : NamesRegistry
        A collection of CSDMS Standard Names.
    newline : str, optional
        Newline character to use for output.

    Returns
    -------
    str
        The SQL commands to create the database.

    Examples
    --------
    >>> import standard_names as csn
    >>> names = csn.NamesRegistry(None)
    >>> names.add('air__temperature')
    >>> print(csn.cmd.snsql.as_sql_commands(names, newline='\\n'))
    ...     # doctest: +REPORT_UDIFF
    BEGIN TRANSACTION;
    CREATE TABLE names (
        id       integer primary key,
        name     text,
        unique(name)
    );
    INSERT INTO "names" VALUES(1,'air__temperature');
    CREATE TABLE objects (
        id       integer primary key,
        name     text,
        unique(name)
    );
    INSERT INTO "objects" VALUES(1,'air');
    CREATE TABLE operators (
        id       integer primary key,
        name     text,
        unique(name)
    );
    CREATE TABLE quantities (
        id       integer primary key,
        name     text,
        unique(name)
    );
    INSERT INTO "quantities" VALUES(1,'temperature');
    COMMIT;
    """
    newline = newline or os.linesep

    from contextlib import closing
    from sqlite3 import connect

    with closing(connect(":memory:")) as db:
        db.cursor().executescript(_NAMES_SCHEMA)

        c = db.cursor()

        for name in names.names:
            c.execute("INSERT INTO names(name) VALUES ('%s');" % name)
        for name in names.objects:
            c.execute("INSERT INTO objects(name) VALUES ('%s');" % name)
        for name in names.quantities:
            c.execute("INSERT INTO quantities(name) VALUES ('%s');" % name)
        for name in names.operators:
            c.execute("INSERT INTO operators(name) VALUES ('%s');" % name)
        db.commit()

        commands = newline.join(db.iterdump())

    return commands


def main():
    """
    Build a database of CSDMS standard names from a list.
    """
    import argparse

    parser = argparse.ArgumentParser(
        description="Build an sqlite database from a list of names"
    )
    parser.add_argument(
        "file", nargs="+", type=argparse.FileType("r"), help="List of names"
    )
    args = parser.parse_args()

    names = NamesRegistry(args.file)
    return as_sql_commands(names)


def run():
    print(main())



#! /usr/bin/env python
"""A CSDMS standard name."""
import re

from six import string_types

from .error import BadNameError


_PREFIX_REGEX = "^[a-z]([a-zA-Z0-9~-]|_(?!_))*"
_SUFFIX_REGEX = "[a-z0-9]([a-z0-9~-]|_(?!_))*[a-z0-9]$"
STANDARD_NAME_REGEX = re.compile(_PREFIX_REGEX + "(__)" + _SUFFIX_REGEX)
#    '^[a-z][a-z0-9_]*[a-z0-9](__)[a-z0-9][a-z0-9_]*[a-z0-9]$'


def is_valid_name(name):
    """Check if a string is a valid standard name.

    Parameters
    ----------
    name : str
        Standard name as a string

    Returns
    -------
    bool
        ``True`` if the string is a valid standard name
    """
    return bool(STANDARD_NAME_REGEX.match(name))


class StandardName(object):

    """A CSDMS standard name.

    Examples
    --------
    >>> import standard_names as csn
    >>> name = csn.StandardName('air__temperature')

    >>> name == 'air__temperature'
    True
    >>> repr(name)
    "StandardName('air__temperature')"

    >>> name.object = 'water'
    >>> repr(name)
    "StandardName('water__temperature')"

    >>> name.quantity = 'density'
    >>> repr(name)
    "StandardName('water__density')"

    >>> name.operators = ('max', )
    >>> repr(name)
    "StandardName('water__max_of_density')"

    >>> name.operators = 'min'
    >>> repr(name)
    "StandardName('water__min_of_density')"
    """

    re = _PREFIX_REGEX + "(__)" + _SUFFIX_REGEX

    def __init__(self, name):
        """Create a standard name object from a string.
        
        Parameters
        ----------
        name : str
            A Standard Name.
        """
        if not is_valid_name(name):
            raise BadNameError(name)

        self._name = name

        (self._object, self._quantity, self._operators) = StandardName.decompose_name(
            name
        )

    @staticmethod
    def decompose_name(name):
        """Decompose a name into its parts.

        Decompose the *name* standard name string into it's constituent
        parts (object, quantity, and operator). Returns a tuple of
        (object, quantity, operator) where object and quantitiy are strings,
        and operator is itself a tuple of strings (or empty).

        Parameters
        ----------
        name : str
            A CSDMS Standard Name.

        Returns
        -------
        tuple of str
            The parts of a name as ``(object, quantity, operators)``

        Examples
        --------
        >>> import standard_names as csn
        >>> name = 'atmosphere_air__elevation_angle_of_gradient_of_temperature'
        >>> csn.StandardName.decompose_name(name)
        ('atmosphere_air', 'temperature', ('elevation_angle', 'gradient'))

        >>> try:
        ...     StandardName.decompose_name('air_temperature')
        ... except BadNameError:
        ...     pass
        """
        try:
            (object_part, quantity_clause) = name.split("__")
        except ValueError:
            raise BadNameError(name)

        (operators, quantity_part) = StandardName.decompose_quantity(quantity_clause)

        return object_part, quantity_part, operators

    @staticmethod
    def compose_name(object, quantity, operators=()):
        """Create a string from the parts of StandardName.

        Parameters
        ----------
        object : str
            An StandardName object.
        quantity : str
            An StandardName quantity.
        operators : iterable of str, optional
            Operators applied to the quantity.

        Returns
        -------
        str
            The standard name composed of the given elements.

        Examples
        --------
        >>> import standard_names as csn
        >>> name = 'atmosphere_air__elevation_angle_of_gradient_of_temperature'
        >>> parts = csn.StandardName.decompose_name(name)
        >>> csn.StandardName.compose_name(*parts) == name
        True
        >>> csn.StandardName.compose_name('air', 'temperature')
        'air__temperature'
        """
        operator = "_of_".join(operators)
        if len(operator) > 0:
            quantity = "_of_".join([operator, quantity])

        return "__".join((object, quantity))

    @staticmethod
    def decompose_quantity(quantity_clause):
        """Decompose a quantity into operators and quantities.

        Decompose the *quantity_clause* string into operator and base
        quantity constituents. Because multiple operators can act on a
        quantity, the operators are given as a tuple regardless of the
        number of operators actually present. Returns the parts of the
        quantity as a tuple of (operators, base_quantity)

        Parameters
        ----------
        quantity_clause : str
            A CSDMS Standard Name quantity.

        Returns
        -------
        tuple of str
            The parts of the quantity as ``(operators, quantity)``.
        """
        quantity_parts = quantity_clause.split("_of_")
        quantity = quantity_parts[-1]
        operators = tuple(quantity_parts[:-1])

        return operators, quantity

    @property
    def name(self):
        """The full standard name as a string."""
        return self._name

    @property
    def object(self):
        """The object part of the standard name."""
        return self._object

    @object.setter
    def object(self, value):
        self._object = value
        self._name = StandardName.compose_name(
            self.object, self.quantity, self.operators
        )

    @property
    def quantity(self):
        """The quantity part of the standard name."""
        return self._quantity

    @quantity.setter
    def quantity(self, value):
        self._quantity = value
        self._name = StandardName.compose_name(
            self.object, self.quantity, self.operators
        )

    @property
    def operators(self):
        """The operator part of the standard name."""
        return self._operators

    @operators.setter
    def operators(self, value):
        if isinstance(value, string_types):
            value = (value,)
        self._operators = value
        self._name = StandardName.compose_name(
            self.object, self.quantity, self.operators
        )

    def __repr__(self):
        return "StandardName(%r)" % self.name

    def __str__(self):
        return self.name

    def __eq__(self, that):
        return self.name == str(that)

    def __ne__(self, that):
        return self.name != str(that)

    def __lt__(self, that):
        return self.name < str(that)

    def __gt__(self, that):
        return self.name > str(that)

    def __cmp__(self, that):
        return self.name == str(that)

    def __hash__(self):
        return hash(self.name)



#! /usr/bin/env python
"""
Example usage:
    sndump -n -o -q -op --format=wiki > standard_names.wiki
"""
from __future__ import print_function

import os
import sys
import argparse

from ..registry import NamesRegistry
from ..utilities import FORMATTERS


_FORMATS = FORMATTERS.keys()


def sndump(file=None, format="plain", sorted=True, keys=None, newline=None):
    """Dump a registry to different formats.

    Parameters
    ----------
    file : str, optional
        Name of a registry file of names.
    format : {'plain'}, optional
        Output format.
    sorted : bool, optional
        Sort results.
    keys : {'names, 'objects', 'quantities', 'operators'} or iterable
        Standard Name element or elements to print.
    newline : str, optional
        Specify the newline character to use for output.

    Examples
    --------
    >>> from __future__ import print_function
    >>> import os
    >>> from six.moves import StringIO
    >>> import standard_names as csn

    >>> lines = os.linesep.join(['air__temperature', 'water__temperature'])
    >>> names = StringIO(lines)

    >>> print(csn.cmd.sndump.sndump(names, newline='\\n'))
    ...     # doctest: +REPORT_NDIFF
    air__temperature
    water__temperature
    """
    newline = newline or os.linesep
    keys = keys or ("names",)
    if file:
        args = (file,)
    else:
        args = ()

    names = NamesRegistry(*args)

    lines = []
    formatter = FORMATTERS[format]
    for key in keys:
        list_to_print = getattr(names, key)
        lines.append(
            formatter(
                list_to_print, sorted=sorted, heading=key, level=2, newline=newline
            )
        )

    return newline.join(lines)


class CustomAction(argparse.Action):
    """Keep track of the order of options are given on the command line."""

    def __call__(self, parser, namespace, values, option_string=None):
        if not "ordered_args" in namespace:
            setattr(namespace, "ordered_args", [])
        previous = namespace.ordered_args
        previous.append(self.dest)
        setattr(namespace, "ordered_args", previous)


def main(args=None):
    """Dump a list of known standard names.

    Parameters
    ----------
    args : iterable of str, optional
        Arguments to pass to *parse_args*. If not provided, use ``sys.argv``.

    Examples
    --------
    >>> import os
    >>> import standard_names as csn
    >>> (fname, _) = csn.registry._get_latest_names_file()
    >>> registry = csn.NamesRegistry()

    >>> names = csn.cmd.sndump.main(['-n', fname]).split(os.linesep)
    >>> len(names) == len(registry)
    True

    >>> objects = csn.cmd.sndump.main(['-o', fname]).split(os.linesep)
    >>> len(objects) == len(registry.objects)
    True

    >>> names = csn.cmd.sndump.main(['-n', '-o', fname]).split(os.linesep)
    >>> len(names) == len(registry) + len(registry.objects)
    True
    """
    parser = argparse.ArgumentParser("Dump known standard names")

    parser.add_argument(
        "-n", nargs=0, dest="names", help="Print standard names", action=CustomAction
    )
    parser.add_argument(
        "-o",
        nargs=0,
        dest="objects",
        help="Print standard objects",
        action=CustomAction,
    )
    parser.add_argument(
        "-q",
        nargs=0,
        dest="quantities",
        help="Print standard quantities",
        action=CustomAction,
    )
    parser.add_argument(
        "-op",
        nargs=0,
        dest="operators",
        help="Print standard operators",
        action=CustomAction,
    )

    parser.add_argument(
        "file", type=argparse.FileType("r"), default=None, help="Read names from a file"
    )
    parser.add_argument(
        "--unsorted", action="store_true", default=False, help="Do not sort names"
    )
    parser.add_argument(
        "--format", choices=_FORMATS, default="plain", help="Output format"
    )

    if args is None:
        args = parser.parse_args()
    else:
        args = parser.parse_args(args)

    try:
        keys = args.ordered_args
    except AttributeError:
        keys = ["names"]

    return sndump(
        file=args.file, format=args.format, sorted=not args.unsorted, keys=keys
    )


def run():
    print(main())



#! /usr/bin/env python
"""Validate a list of names."""
from __future__ import print_function

import os
import sys
import argparse

from ..registry import NamesRegistry
from ..error import BadRegistryError


def main(args=None):
    """Validate a list of names.

    Examples
    --------
    >>> from __future__ import print_function
    >>> import os
    >>> import standard_names as csn

    >>> (fname, _) = csn.registry._get_latest_names_file()
    >>> csn.cmd.snvalidate.main([fname])
    0

    >>> import tempfile
    >>> (fd, fname) = tempfile.mkstemp()
    >>> os.close(fd)

    >>> with open(fname, 'w') as fp:
    ...     print('air__temperature', file=fp)
    ...     print('Water__temperature', file=fp)
    ...     print('water_temperature', file=fp)

    >>> csn.cmd.snvalidate.main([fp.name])
    2

    >>> os.remove(fname)
    """
    parser = argparse.ArgumentParser("Validate a list of standard names")

    parser.add_argument(
        "file",
        type=argparse.FileType("r"),
        nargs="+",
        default=None,
        help="Read names from a file",
    )

    if args is None:
        args = parser.parse_args()
    else:
        args = parser.parse_args(args)

    try:
        names = NamesRegistry(args.file)
    except BadRegistryError as err:
        print(os.linesep.join(err.names), file=sys.stderr)
        return len(err.names)
    else:
        return 0


def run():
    sys.exit(main())
